/*
 * Copyright 1999 Pavel Machek, distribute under GPL version 2 or later.
 *
 * This converts interrupt into signal
 */

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/malloc.h>
#include <linux/init.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/wait.h>
#include <linux/fs.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/kbd_kern.h>
#include <linux/pci.h>

#include <net/irda/irda_device.h>

#include <asm/io.h>
#include <asm/dma.h>
#include <asm/byteorder.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/irq.h>

#define ID printk( "n2cd: " )
#define MAXDEVS 64

#define DEBUG_OUTPUT 0

MODULE_AUTHOR("Pavel Machek <pavel@suse.cz>");
static int pid = 0;
MODULE_PARM(pid, "i");
static int irq = 3;
MODULE_PARM(irq, "i");

static void enable_now_do(unsigned long unused)
{
#if DEBUG_OUTPUT == 1
	printk( "enough time, enabling interrupt..." );
#endif
	enable_irq(irq);
#if DEBUG_OUTPUT == 1
	printk( "enabled\n" );
#endif
}

static struct timer_list enable_now_timer = {NULL, NULL, 0, 0, enable_now_do};

static void
interrupt( int irq, void *dev_id, struct pt_regs *regs )
{
#if DEBUG_OUTPUT == 1
	printk( "Interrupt %d came...", irq );
#endif
	kill_proc(pid, SIGUSR1, 1);
	disable_irq(irq);
        enable_now_timer.expires = jiffies + HZ;
        enable_now_timer.prev = enable_now_timer.next = NULL;
        add_timer(&enable_now_timer);
#if DEBUG_OUTPUT == 1
	printk( "ok\n" );
#endif
}

#ifdef MODULE
int init_module(void)
{
	printk( "i2sig: V0.1.0, copyright 1999 Pavel Machek\n" );

        if (request_irq( irq, interrupt, SA_SHIRQ, "i2sig", (void *) &irq)) {
                printk( "Can not get IRQ\n" );
                return -EAGAIN;
        }
	printk( "Success." );
	return 0;
}

void cleanup_module(void)
{
	free_irq(irq, (void *) &irq);
}
#endif




