/*
 * Port IO routines + emulation of low level layers from ltmodem.sys.
 *
 *  Copyright (c) 1999 Richard J.M. Close
 *  Copyright (c) 1999 Pavel Machek <pavel@suse.cz>
 *  Copyright (c) 1999 Jamie Lokier
 *
 *  Can be freely distributed and used under the terms of the GNU GPL.
 */

#include "portIO.h"


// IO port subroutines etc.
// The first section has routines based on the NT HAL library calls.
// The implementation in this file is not a copy of the real code but
// has the same interface and is designed to be used by user space processes.
// A version for a device driver will be implemented at a later date, this
// will be much simpler as more system resources are available to drivers.

// Also note that the pausing versions of inb, inw etc. (inb_p inw_p etc.)
// have been used.

// Low level interface
//=============================================================================
//
// The following section contains routines that have been "reverse engineered" from those in
// in the NT driver code. The functions are named according to thier names in original
// lucent .sys driver.


// Port variables - these are initialised by port_io_init. The values below are purely
// for illustration purposes.
// PCIOutAddr holds a (?) base port address and PCIInAddr has this base + 4.
unsigned int PCIOutAddr = 0x6200;
unsigned int PCIInAddr = 0x6204;

// I think this initial value is wrong. I need to find out how this is being set.
// In read_Blacktip_Resource it is set to 118h, 110h, 108h, or 100h. This may be a clue!
// dp_run_rom sets it to 260h.
unsigned int BaseAddress;
 
unsigned int BaseAddress2; 
unsigned int ComAddress = 0;
unsigned int word_59ED7 = 0;

// DSP RAM read fail count.
unsigned int dp_readram_failcnt = 0;


void Init_Mars(void)
{
  Write_mdm_byte(0x37, 0x22);
  Write_mdm_word(0x30,    1);
  Write_mdm_word(0x3c, 0xff);
  Write_mdm_word(0x3e,	  1);
}

// Main initialisation routine called from ltmodem.c code.
void port_io_init(unsigned int Portbase1, unsigned int Portbase2)
{
  // Which portbase do we use? Pavel used the second so ...
  BaseAddress = Portbase2;
  BaseAddress2 = BaseAddress + 2;

  printf( "Initialising with ports at %x and %x, ",
          Portbase1, Portbase2 );
  printf( "assuming portbase at %x.\n", BaseAddress );

  // Set up other portbases.
  PCIOutAddr = Portbase1;
  PCIInAddr = Portbase1 + 4;

  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
    perror("portIO: iopl(3) failed");
    fprintf(stderr, "This program must be run as root.\n");
  }
}

// Places args in a long word to go to port, then reads back.
unsigned int PciRead_dword (unsigned char arg1, unsigned char arg2, unsigned char arg3)
{
  unsigned int ecx;
  unsigned char al;

  // shl al, 03h  (al = arg1). shift 3 left, ie. mult by 8.
  // and cl, 07h (cl = arg3).  mask off all but last 3 bits.
  // or al, cl                 OR the results.
  al = (arg1 * 8) | (arg3 & 7);

  //xor ecx, ecx -> ecx = 0 !
  //mov ch, al                 set byte 1 to above result.
  //mov  cl , byte ptr [ esp + 8 ]
  ecx = (256 * al) + arg2;         // put second arg in byte0. 
  
  // or ecx , 080000000h    Set most sig bit.
  ecx = ecx | 0x80000000; 

#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outw_p (ecx, PCIOutAddr);
  //WRITE_PORT_ULONG (PCIOutAddr, ecx);
 
  return (inw_p (PCIInAddr));
  //	ret 0Ch			;0x00025425 :	c20c00 wind esp back 12 bytes.
}

unsigned int PciRead_word (unsigned char arg1, unsigned char arg2, unsigned char arg3)
{
  unsigned int dataOut;
  unsigned short readPort;

  dataOut = (256 * (arg1 * 8) | (arg3 & 7)) + arg3;
  dataOut = dataOut | 0x80000000; 
  
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outw_p (dataOut, PCIOutAddr);

  readPort = PCIInAddr | (arg2 & 2);
  return (inb_p (readPort));
}

void PciWrite_dword (unsigned int arg1, unsigned int arg2, unsigned int* ptrValue)
{
  unsigned int eax, ecx, Value;

  ecx = arg1 * 2084; // shl ecx, 0xBh
  ecx = ecx & 0x0000ffff;  //movzx ecx, cx

  eax = arg2 & 0xFF;
  eax = eax | ecx;
  eax = eax | 0x80000000;
   
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outw_p (eax, PCIOutAddr);
   
  Value = *ptrValue;
  outw_p (Value, PCIInAddr);
}

unsigned char Read_mdm_byte (unsigned int offset)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  return inb_p (BaseAddress + offset);
}

// Place value in port (base address in BaseAddress) plus offset.
void Write_mdm_byte (unsigned int portOffset, unsigned char value)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outb_p (value, (BaseAddress + portOffset));
}

// Place value in port (base address in BaseAddress) plus offset.
void Write_mdm_word (unsigned int portOffset, unsigned short value)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outw_p (value, (BaseAddress + portOffset));
}

void dp_out_port (int arg)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif
  outb_p (BaseValue, BaseAddressIndex);
  outb_p (arg, BaseAddressData);
}

int dp_in_port(void)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif
  outb_p (BaseValue, BaseAddressIndex);
  return inb_p (BaseAddressData);
}

//
// If you want to monitor communications with ltmodem, this is quite a good place 
//
void dp_regwrite (unsigned int port, unsigned char val)
{
#if LT_DEBUG_IO
  printf("\n{%x,%x}\n", port, val );
#endif
  BaseValue = port;
  BaseAddressIndex = BaseAddress;
  dp_out_port(val);
}

unsigned int dp_regread (unsigned int port)
{
  BaseValue = port;
  if (port <= 0x7f) {
    BaseAddressIndex = BaseAddress;
    return dp_in_port();
  }
  if (port < 0xa0) {
    BaseAddressIndex = BaseAddress2;
    return dp_in_port();
  }
  if (port <= 0xcf) {
    BaseAddressIndex = BaseAddress;
    return dp_in_port();
  }
  BaseAddressIndex = BaseAddress2;
  return dp_in_port();
}

void dp_int_regwrite(unsigned int port, unsigned int val)
{
#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif
  outb_p(port, BaseAddress);
  outb_p(val, BaseAddressData);
}

// 656 version uses BaseAddress2!
int dp_int_regread(unsigned int port)
{

  unsigned short outAddress;

  if (port <= 0x7f) 
    outAddress = BaseAddress;
  else if (port < 0xa0) 
    outAddress = BaseAddress2;
  else if (port <= 0xcf)
    outAddress = BaseAddress;
  else
    outAddress = BaseAddress2;

#if TRY_INTERRUPTS
  // Grab access to all the IO ports.
  if (iopl(3) < 0) {
	perror("portIO: iopl(3) failed");
	fprintf(stderr, "This program must be run as root.\n");
  }
#endif

  outb((unsigned char)port, outAddress);

  return inb(BaseAddressData);
}


int dp_dsp_regread (unsigned int arg)
{
  dp_byte_f = 0;
  dp_regwrite(0xD8, 0x18);
  dp_regwrite(0x36, arg);
  dp_regwrite(0x37, 0x06);
  dp_cmd_timer = x_current_time();
  while(1) {
    if (!dp_byte_f) {
      if (x_elapsed_time(dp_cmd_timer) < 0xc8)
	     continue;
    }
    if (dp_byte_f != 2) {
      dp_regwrite(0xD8, 0xFF);
      dp_byte_e = 0;
    }
    //printf("dp_dsp_regread: exiting with dp_byte_f = %d.\n", dp_byte_f);
    return dp_byte_e;
  }
}

int dp_read_dsp_ram(int where)
{
  dp_byte_f = 0;
  dp_regwrite(0xd8, 0x18);
  dp_regwrite(0x34, where);
  dp_regwrite(0x35, (where & 0xff00) / 256);
  dp_regwrite(0x37, 4);
  dp_cmd_timer = x_current_time();

  do {
    if (dp_byte_f != 0)
      break;
  } while (x_elapsed_time(dp_cmd_timer) < 0xc8);

  if (dp_byte_f != 1) {
    dp_regwrite(0xd8, 0xff);
    dp_byte_e = 0;
    dp_byte_d = 0;
    if (x_dsp_mars)
      Write_mdm_word(0x3c, 0x1ff);

	// Increment fail count.
    dp_readram_failcnt++;
  }
  else
    // Clear fail count.
    dp_readram_failcnt = 0;

  //printf("dp_read_dsp_ram: exiting with dp_byte_f = %d.\n", dp_byte_f);
  return ((dp_byte_e * 256) + dp_byte_d);
}

void
dp_write_dsp_ram(int addr, int val)
{
  dp_regwrite(0xd8, 0x18);
  dp_regwrite(0x32, val & 0xff);
  dp_regwrite(0x33, (val>>8) & 0xff);
  dp_regwrite(0x34, addr & 0xff);
  dp_regwrite(0x35, (addr>>8) & 0xff);
  dp_regwrite(0x37, 1);
  wait_for_core_read();
}

// This should be correct now.
void dp_modem_command (unsigned int command, unsigned int a4, unsigned int a8)
{

  if ((command == 0x0E) || (command == 0x0C)) 
	{
	  if ((dp_dsp_regread(0xB5) & 1) == 0)
		dp_regread(0xB0);
	  
	  if ((dp_dsp_regread(0xB5) & 0x10) == 0)
		{
		  printf("modem command: something active at 0xb5?\n");
		  dp_regwrite(0xB6, 0xEF);
		  dp_dsp_regread(0xB0);
		  dp_regwrite(0xB0, 0);
		}
 
	  dp_regwrite(0xB7, 0xFF);
	  dp_regwrite(0xB6, 0xEE);
	  dp_bamil_rd7 = 0xF3;
	  dp_regwrite(0xD7, 0xF3);
	  dp_dsp_data_in_progress = 0;
	}

  dp_regwrite(0xD8, 0x18);
  dp_regwrite(0x35, a8);
  dp_regwrite(0x36, a4);
  dp_regwrite(0x37, command);
  return wait_for_core_read();
}

void dp_modem_command_long(int a0, int a4, int a8, int aC, int a10)
{
  dp_regwrite(0xD8, 0x18);
  dp_regwrite(0x33, a10);
  dp_regwrite(0x34, aC);
  dp_regwrite(0x35, a8);
  dp_regwrite(0x36, a4);
  dp_regwrite(0x37, a0);
  return wait_for_core_read();
}

void x_output_init(void)
{
  if (x_dsp_mars) {
    dp_regwrite(0xb8, 1);
    dp_regwrite(0xb9, 1);
    dp_regwrite(0xba, 0x40);
    dp_regwrite(0xbb, 0x40);
    dp_regwrite(0xdb, 0x41);
    dp_regwrite(0xdc, 0);
  } else {
    dp_regwrite(0xb8, 0xe0);
    dp_regwrite(0xb9, 0xc0);
    dp_regwrite(0xba, 0x80);
    dp_regwrite(0xbb, 0x80);
    dp_regwrite(0xdb, 0);
    dp_regwrite(0xdc, 0);
  }
}


void x_output(int arg)
{

// Start of main switch. (loc_2e6f1)
repeat:
switch (arg)
  {
  case 0: // 0002e71c
	x_output_init ();
	set_boardid ();
	x_set_hardware_options ();
	break;

  // loc_2e730
  case 1:
    x_output_deinit();
    break;

  // loc_2e73a
  case 2:
    dp_modem_command(0x11, 0, 0);
    break;

  // loc_2e740
  case 3:
    dp_modem_command(0x12, 0, 0);
    break;

  case 4: case 5: case 6:
    break;

  // loc_2e706
  case 7:
	dp_modem_command (0x27, 1, 0);
	arg = 24;
	goto repeat;

  // loc_2e6f8
  case 8:
    dp_modem_command(0x27, 2, 0);
    arg = 23;;
    goto repeat;

  case 9: // 0002e7dc
	if (x_dsp_mars
		&& ((S(0x7e) & 7) == 1 || (S(0x7e) & 7) == 2))
	  dp_write_dsp_ram (0x0d, 0x60);
	else
	  dp_write_dsp_ram (0x0d, 0x40);
	break;

  case 10: // 0002e7b8
	if (x_dsp_mars
		&& ((S(0x7e) & 7) == 1 || (S(0x7e) & 7) == 2))
	  dp_write_dsp_ram (0x0d, 0x100);
	else
	  dp_write_dsp_ram (0x0d, 0x70);
	break;


  // loc_2E7B1
  case 11:
    dp_write_dsp_ram(0x0d, 0x130);
    break;

  // loc_2E7AA
  case 12:
    dp_write_dsp_ram(0x0d, 0x400);
    return;

  case 13: // 0002e74e
	if (H_0x7f & 0x80)
	  dp_change_mercury_gain (0, 2, H_mercury_ciocb_MSB_tx_gain,
							  1 | H_mercury_ciocb_LSB_BOM_control);
	else
	  writebio (x_dsp_mars ? 0x01 : 0x40, 0);
	break;

  case 14: // 0002e77e
	if (H_0x7f & 0x80)
	  dp_set_mercury_gain ();
	else
	  writebio (x_dsp_mars ? 0x01 : 0x40, 1);
	break;

  case 15: // 0002e9e4
	if (H_shunt_relay_on_time != 0 && !CpqFlag)
	  writebio (x_dsp_mars ? 0x304 : 0x80, 0);
	break;

  case 16: // 0002ea07
	if (H_shunt_relay_on_time != 0 && !CpqFlag)
	  writebio (x_dsp_mars ? 0x304 : 0x80, 1);
	break;

  case 17: // 0002ea34
	writebio (0xffff, 0);
	if (CpqFlag)
	  writebio (0x302, 0);
	break;

  case 18: // 0002ea55
	writebio (0xffff, 1);
	if (CpqFlag)
	  writebio (0x302, 1);
	break;

  case 19: case 20: case 21: case 22:
	break;

  case 23: // 0002e807
	if (x_dsp_mars && ((S(0x7e) & 7) == 1 || ((S(0x7e)) & 7) == 2))
	  {
		if (S_0x81 == 0x1179)
		  writebio (0x1240, 0);
	  }
	else
	  {
		dp_regwrite (0xda, x_dsp_mars ? 5 : 7);
		dp_regandor (0xcb, 0xf3, 0x00);
		if (x_dsp_mars)
		  init_1027 (1);
		writebio (x_dsp_mars ? 0x1240 : 0xffff, 0);
		if (CpqFlag)
		  writebio (0x180, 0);
	  }
	break;

  case 24: // 0002e89b
	if (x_dsp_mars && ((S(0x7e) & 7) == 1 || ((S(0x7e)) & 7) == 2))
	  {
		if (S_0x81 == 0x1179)
		  writebio (0x1240, 1);
		if (CpqFlag)
		  writebio (0x180, 1);
	  }
	else
	  {
		writebio (x_dsp_mars ? 0x1240 : 0xffff, 1);
		if (CpqFlag)
		  writebio (0x180, 1);
		if (x_dsp_mars)
		  {
			dp_regandor (0xcb, 0xf3, 0x00);
			init_1027 (0);
		  }
		dp_regandor (0xcf, 0x3f, 0x00);
		dp_regwrite (0xda, 1);
		dp_regandor (0xcb, 0xff, 0x08);
		dp_modem_command_long (0x03, 0xd1, 0, 0xea, 0);
	  }
	break;

// loc_2EACF
  case 25:
    if (x_dsp_mars)
      writebio(0xffff, 1);
    else
      writebio(0x180, 1);
    break;

// loc_2EAC3
  case 26:
    if (x_dsp_mars)
      writebio(0xffff, 0);
    else
      writebio(0x180, 0);
    break;

  // loc_2EA72
  case 27:
    if (x_dsp_mars)
       writebio(0x308, 0);
    else
       writebio(0x304, 0);
    break;

  // loc_2EA7C
  case 28:
    if (x_dsp_mars)
       writebio(0x308, 1);
    else
       writebio(0x304, 1);
    break;

  // loc_2EA8D
  case 29:
    if (!x_dsp_mars)
       writebio(0x308, 0);
    else
       writebio(0xffff, 0);
    return;

  // loc_2EA97
  case 30:
    if (!x_dsp_mars)
       writebio(0x308, 1);
    else
       writebio(0xffff, 1);
    break;

  // loc_2EAA8
  case 31:
    if (!x_dsp_mars)
       writebio(0x302, 0);
    else
       writebio(0xffff, 0);
    break;

  // loc_2EAB2
  case 32:
    if (!x_dsp_mars)
       writebio(0x302, 1);
    else
       writebio(0xffff, 1);
    break;

  case 33: case 34: case 37: case 38:
	break;

  case 35: // 0002e9ba
	writebio (0x1001, 0);
	break;
  case 36: // 0002e9bd
	writebio (0x1001, 1);
	break;
  case 39: // 0002e9c8
	writebio (0x1004, 0);
	break;
  case 40: // 0002e9cb
	writebio (0x1004, 1);
	break;
  case 41: // 0002e9d9
	writebio (0x1008, 0);
	break;
  case 42: // 0002e9d6
	writebio (0x1008, 1);
	break;
  case 43: // 0002e964
	if (x_dsp_mars)
	  {
		dp_regwrite (0xb9, dp_regread (0xb9) & 0xf0);
		dp_regwrite (0xb8, 0x0d);
		dp_regwrite (0xb9, dp_regread (0xb9) & 0xf0);
	  }
	break;

  case 44: // 0002e99d
	if (x_dsp_mars)
	  dp_regwrite (0xb8, 0x01);
	break;

  default: die( "x_output: %d - not implemented\n", arg );
  }
}


// Needs finishing.
int x_input(int arg)
{
  unsigned char al;

  switch (arg) {
  case 0x38:   // loc_2E6A2
    if (!x_dsp_mars)
	  al = readbio(0x1120);
    else
      al = readbio(0xffff);
	if (al != 0) return 1;

    if (!x_dsp_mars)
	  readbio(0x1120);
    else
      readbio(0xffff);
	return 0;

  case 0x39: // loc_2E67D
    if (!x_dsp_mars)
	  al = readbio(0x1120);
    else
      al = readbio(0xffff);
	if (al == 0) return 1;

    if (!x_dsp_mars)
	  readbio(0x1120);
    else
      readbio(0xffff);
	return 0;

  case 0x3e: // loc_2E671
	return 0;

  case 0x40: // loc_2E667
    printf( "x_input: error 0x%x - not implemented!\n", arg );
    break;

  case 0x41: 
    if (cell_active == 1) return 0; // loc_2E6D1
    if (!x_dsp_mars) {
       al = readbio(0x1140);
       if (al != 0) {
          al = readbio(0x1140);
          return 0;
	   }
	   else {  // loc_2E6D7
          return 1;
	   }
	}
	else {     
       al = readbio(0x1110);
       if (!al) return 1;
       readbio(0x1110);
       return 0;
	}

  default: /* 6d1 */
	return 0;
  }
  return 0;
}


void dp_regandor (unsigned int port, unsigned int val1, unsigned int val2)
{

   dp_regwrite (port, ((dp_regread (port) & val1) | val2));

}

//========================================================================
//
//

void dp_board_enable (void)
{

  if (CpqFlag== 1) {
	// Compaq specific stuff perhaps?
    if (ComAddress == BaseAddress) {

       if ((word_59ED7 != 0xCCB0) && (word_59ED7 != 0xCDB0))
          BaseAddress = 0xEC; 
	   else
          BaseAddress = 0xBC;
    }
    else if ((BaseAddress != 0xEC) && (BaseAddress != 0xBC))
       BaseAddress = BaseAddress & 0xFFF8;
  }

  x_chip_version = 0;

  BaseAddressIndex = BaseAddress;

  // ToshibaFlag = 1; // I don't think so!

  BaseAddressData = BaseAddress + 1;

  BaseAddress2 = BaseAddress + 2;

  Init_Mars();

  dp_regwrite (0xCB, 8);

  modem_enabled = 1;

  if (x_dsp_mars == 1) return;

  // I think my card has a mars chipset so this stuff
  // probably doesn't matter (yet).

  //dp_dsp_regread_poll(6, 0xC0, 2);
  // etc...

}

// Based on version 568.
void dp_board_disable (void)
{
  //if(S(0x81) == 0x12a3)
  //	SetPCI_Conf_Byte(0x12a3, 0x48, 0, 1); ??

  dp_regwrite(0xd6, 0xf);
  dp_regandor(0xcd, 0xff, 0x30);

  if (x_dsp_mars != 1) return;

  // Not a Mars chipset.
  Write_mdm_word(0x3c, 0x1ff);
  Write_mdm_word(0x3e, 0);
}

/* Ugly, feel free to rewrite it to be really readable. But I do not care - this function is not critical in any way */
int
dp_dsp_rom_checksum(void)
{
  int i;
  dp_byte_f = 0;
  dp_modem_command(0x10, 0, 0);
 l94E:
  if (dp_byte_f == 5)
    goto l972;
  i = x_elapsed_time(dp_cmd_timer);
  if (i<64)
    goto l94E;
  if (dp_byte_f != 5)
    goto l987;
 l972:
  dp_version = dp_byte_e << 8 | dp_byte_d;
  goto l98E;

 l987:
  dp_version = 0;
 l98E:
  dp_byte_f = 0;
  dp_modem_command(0x14, 0, 0);

 l99D:
  if (dp_byte_f == 3)
    goto l9C1;
  i = x_elapsed_time(dp_cmd_timer);
  if (i < 0xc8)
    goto l99D;
  if (dp_byte_f != 3)
    goto l9CA;

 l9C1:
  return dp_read_dsp_ram(0x49);

 l9CA:
  printf( "Returning zero, checksum failed?\n" );
  return 0;
}

int
dp_change_mercury_gain(int a0, int a4, int a8, int aC)
{
  int res;
  //printf( "Change mercury gain: %x, %x, %x, %x =", a0, a4, a8, aC );
  dp_regwrite(0x32, aC);
  dp_modem_command_long(0x2f, 1, a0, a4, a8);
  res = dp_read_dsp_ram(0x3a);
  //printf( "%x\n", res );
  return res;
}

void
dp_set_mercury_gain(void)
{
  dp_change_mercury_gain(homol[0xF30-HBASE],
						 homol[0xF31-HBASE], 
						 homol[0xF32-HBASE],
						 homol[0xF33-HBASE]);
}

int
readbio(int arg)
{
  int var, res;
  if ((arg & 0xf00) == 0xf00)
    return 0;
  if ((arg & 0xf00) == 0x100) 
    var = 0xbb;
  else {
    if ((arg & 0xf00) == 0x200)
      var = 0xdc;
    else
      var = 0xb9;
  }
  res = dp_regread(var) & arg;
#if LT_DEBUG
  printf( "[readbio: %x]", res );
#endif
  return res;
}

void
writebio(int val, int strange)
{
  int ebx = val;
  int addr;

  //printf( "writebio(%x, %x)\n", val, strange);

#if 0
  if ((val & 0xff) > 0x80)
    return;
#endif

  val = (val & 0xffffff00) | 0xb9;
  if ((val & 0xf00) != 0x100)
    goto l482;
  val = (val & 0xffffff00) | 0xbb;

 l441:
  addr = val & 0xff;
  if (strange == 1) {
    if (!(val & 0x1000))
      goto l457;
  }

  if ((strange != 0) || (!(val & 0x1000))) {
    strange = dp_regread(addr) & ((~ebx) & 0xff);
    dp_regwrite(addr, strange);
    return;
  }

 l457:
  strange = dp_regread(addr) | (ebx & 0xff);
  dp_regwrite(addr, strange);
  return;

 l482:
  if ((val & 0xf00) == 0x200) {
    val = (val & 0xffffff00) | 0xdc;
    goto l441;
  }

  if ((val & 0xf00) != 0x300)
    goto l441;
  writebiom(ebx, strange);
  return;
}

void
writebiom(int val, int strange)
{
  int al;
  al = val & 0x0f;
  if ((strange == 1) && (!(val & 0x1000)))
      goto do_and;

  if ((strange) || (!(val & 0x1000))) {
    homol[0xF33-HBASE] &= ~al;
    dp_set_mercury_gain();
    return;
  }

 do_and:
  homol[0xF33-HBASE] |= al;
  dp_set_mercury_gain();
  return;
}


//=================================================================
   
// This is complete!
void dp_init_local_phone_timer(void)
{
  dp_local_phone_timer = x_current_time();
}

// Probably cellphone related, so not bothered about this.
void cell_init (void)
{
}

// Used by dp_init_variables.
void dp_init_blacklist(void){

  dp_failures = 0;
  dp_blacklist_calls = 0;

  // Clears/sets up some locations around 59A94?
  /*  {
    int ecx = 0;
    int edx;
    int* eax = unk_59A94; // Table start address + 0x14(?)
    for (edx = 0x18; edx>0; edx--){
	  *(eax - 0x14) = 0;  // byte
	  *eax = 0;           // 32 bit dword
	  *(eax + 4) = 0;     // byte
      eax = eax + 0x20;
	}
  } */

}

// Complete!
void dp_init_local_phone_state(void){
  dp_init_local_phone_timer();
  S(0x5A) = 0;
}

void writecioc (unsigned int arg1, unsigned int arg2)
{
   // Read reg D5h until bit 2 is set, or time has exceeded 500 milliseconds.
   unsigned int startTime = x_current_time();
  
   while ((dp_regread(0xd5) & 2) == 0)
	 {
	   if (x_elapsed_time(startTime) >= 0x1f4) break;
	 }

   dp_regwrite(0x2b, arg1); 
   dp_regwrite(0x2a, arg2); 
}

// Empty for now.
void set_boardid(void){}
void x_output_deinit(void){}

// Used to continously monitor port activity.
void continous_monitoring(int monitor_type, int monitoring_interval){

  // Iteration loop variable.
  unsigned int i;

  // Main monitoring loop.
  while (1) {
	// Monitor registers or I/O depending on current setting.
    if (monitor_type & 1) { 
	  if (isapnp_modem)
		show_isahex_dump();
	  else {
		// Re-scan device.
		pci_scan_bus(pacc);
		modem_dev = scan_device(modem_dev->dev);
		// Show the hex.
		show_pcihex_dump(modem_dev);
	  }
	}
    if (monitor_type & 2) {
	  // Get access to all of I/O space.
	   if (iopl(3) < 0) {
		 perror("ltmodem: iopl()");
		 fprintf(stderr, "This program must be run as root.\n");
	   }
	   else {
		 // Read the ports.
         printf("I/O ports\n ");
		 // Loop for each known port.
         for (i=1;i<=io_cnt;i++) {
            unsigned int j;
            printf( "%x: ", io_address[i] );
            for (j=0; j<io_length[i]; j++) {
			   // Read each byte, one at a time.
               printf("%2x ", inb(io_address[i] + j));
	           if ( !((j+1)%16) )
		          printf( "\n%x: ", io_address[i] + j + 1);
	        }
            putchar('\n');
	     }
            // Free the ports.
            iopl(0);
       }
    }
	// Wait for the configured amount of time. 
	usleep(monitoring_interval * 1000);
  }
}

void set_modem_io(void)
 {
   unsigned int i;
   /* value input by user. */
   unsigned int input = 0;
   unsigned int offset = 0;
   /* CR input after first char. */
   char trash; 

   if (io_cnt > 0) {
	 printf("I/O ports found:\n");
	 for(i=1; i<=io_cnt; i++){
	   printf("%d - I/O port at %x.\n", i, io_address[i]);
	 }
     // Select the port.
     printf ("Port to write long word at (in hex): ");
	 scanf ("%x%c", &input, &trash);
     // Write the value.
	 if ((input > 0) && (input <= io_cnt)) {
	   // Allow user to write to selected I/O port.
	   i = input;
       printf ("Write to %x, enter value (in hex): ", io_address[i]);
	   scanf ("%x%c", &input, &trash);

       // Get offset to use.
       printf ("Offset to use (in hex): ");
       scanf ("%x%c", &offset, &trash);

	   printf ("Writing %2x to I/O port %x + %x.\n",
                   input, io_address[i], offset);
	   // Write value to port.
           outb_p  (input, (io_address[i] + offset));

	   ///* Delay 10ms. */
	   usleep(10000);
	   /* Read back the port. */
	   printf("Read back value: %2x.\n", inb_p(io_address[i] + offset));

	 }
   }
 }

void dp_run_rom (void)
{
  unsigned char byte;

  dp_modem_idle_cmd ();

  byte = dp_regread (0xde);
  byte |= 1;
  dp_regwrite (0xde, byte);
  byte &= 0xdf;
  dp_regwrite (0xde, byte);
  x_sleep (2);
  byte &= 0xfe;
  dp_regwrite (0xde, byte);
  x_sleep (2);

  if (CpqFlag == 1)
    {
      unsigned short save = BaseAddress;
      BaseAddress = 0x260;
      dp_regwrite (0xc5, save);
      dp_regwrite (0xc6, save >> 8);
      BaseAddress = save;
    }

  dp_init_modem ();
  dp_regandor (0xda, 0xf7, 0x00);
}


